// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
//
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html

/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////

// Change Log
//
// Date          Version     Developer    Changes
//
// 16 Jan 2003   V 0.0.1.5   Tom Barry	  Use AvisynthPluginInit2

// 25 Nov 2002   V 0.0.1.4   Tom Barry	  Don't filter YV12 Chroma

// 13 Nov 2002   V 0.0.1.3   Tom Barry	  YV12 support 

// ?? Jul 2002   V 0.0.1.2   Tom Barry	  Misc bug fixes  

// 27 Jun 2002   V 0.0.1.0   Tom Barry	  New UnFilter 
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////
/*

		UnFilter is a simple Avisynth Soften/Sharpen filter designed to (slightly) reverse previous
		efforts at softening or edge enhancment that are common in DVD mastering.

		It is invoked as:

			UnFilter(HSharp, VSharp) 
		
		Where both values are integers between -100 to +100.

		Positive values mean Sharpen, negative values mean soften. Zero is neutral.

        Known issues and limitation:
        1) 	Assumes YUV (YUY2) Frame Based input. Use an AVIsynth function to convert first if
	        needed.        

		2)  Currentyl still require the pixel width to be a multiple of 4.

        Temporary file locations:

        For now, see www.trbarry.com/UnFilter.zip for both source and DLL.

		See the comments in UnFilterALL.inc for math and method used.

		THIS PROGRAM IS STILL NEW AND EXPERIMENTAL! I have not yet tested MMX or 3DNOW
		but they may work if I didn't miss any unsupported instructions in cleanup.

  Please send comments to Tom Barry (trbarry@trbarry.com)

*/
#include "internal.h"
//#include "avisynth.h"
#include "UnFilter.h"

extern  "C" BOOL UnFilter_SSE( BYTE* pDest, const BYTE* pCopySrc,
			int rowsize,  int height, int dst_pitch, int src_pitch, int HSharp, int VSharp);

extern  "C" BOOL UnFilter_MMX( BYTE* pDest, const BYTE* pCopySrc,
			int rowsize,  int height, int dst_pitch, int src_pitch, int HSharp, int VSharp);

extern  "C" BOOL UnFilter_3DNOW( BYTE* pDest, const BYTE* pCopySrc,
			int rowsize,  int height, int dst_pitch, int src_pitch, int HSharp, int VSharp);

extern  "C" BOOL UnFilter_SSE_YV12( BYTE* pDest, const BYTE* pCopySrc,
			int rowsize,  int height, int dst_pitch, int src_pitch, int HSharp, int VSharp);

extern  "C" BOOL UnFilter_MMX_YV12( BYTE* pDest, const BYTE* pCopySrc,
			int rowsize,  int height, int dst_pitch, int src_pitch, int HSharp, int VSharp);

extern  "C" BOOL UnFilter_3DNOW_YV12( BYTE* pDest, const BYTE* pCopySrc,
			int rowsize,  int height, int dst_pitch, int src_pitch, int HSharp, int VSharp);


class UnFilter : public GenericVideoFilter {
    int  Max_Frames;

    PVideoFrame src;
    PVideoFrame dst;
    const unsigned char* srcp;
    unsigned char* dstp;
	BOOL ItIsYV12;
    int src_pitch;
    int dst_pitch;
    int rowsize;
    int height;
    int HSharp;
    int VSharp;
	BYTE* oldBuff;
	BYTE* newBuff;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);

public:
    UnFilter(PClip _child, int _HSharp, int _VSharp,
		IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
        Max_Frames = vi.num_frames;
		HSharp = _HSharp;				// Horizontal Sharpness
		VSharp = _VSharp;				// Vertical Sharpness
		SSE2enabled = (env->GetCPUFlags() & CPUF_SSE2) !=0;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;
		_3DNOWenabled = (env->GetCPUFlags() & CPUF_3DNOW) !=0;

	    if (vi.IsYUY2())
		{
			ItIsYV12 = FALSE;
		}
		else if(vi.IsYV12())
		{
			ItIsYV12 = TRUE;
		}
		else
		{
			env->ThrowError("UnFilter: Supports YUY2 and YV12 color formats only");
		}
	    if (vi.width % 2)
		{
			env->ThrowError("UnFilter: Horizontal pixels must divide by 2");
		}
    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
    int __stdcall ProcessFrame();

};


PVideoFrame __stdcall UnFilter::GetFrame(int inFrame, IScriptEnvironment* env) 
{
    src = child->GetFrame(inFrame, env);
    dst = env->NewVideoFrame(vi);
    if (inFrame > Max_Frames-1 || inFrame < 0)
    {
        return dst;
    }

	if (!ItIsYV12)
	{								// do YUY2
		srcp = src->GetReadPtr();
		dstp = dst->GetWritePtr();
		src_pitch = src->GetPitch();
		dst_pitch = dst->GetPitch();
		rowsize = dst->GetRowSize();
		height = dst->GetHeight();      
		ProcessFrame();      
	}
	else							// do YV12
	{
		srcp = src->GetReadPtr(PLANAR_Y);
		dstp = dst->GetWritePtr(PLANAR_Y);
		src_pitch = src->GetPitch(PLANAR_Y);
		dst_pitch = dst->GetPitch(PLANAR_Y);
		rowsize = dst->GetRowSize(PLANAR_Y);  
		height = dst->GetHeight(PLANAR_Y);
		ProcessFrame();       

		// Next, the U plane
		srcp = src->GetReadPtr(PLANAR_U);
		dstp = dst->GetWritePtr(PLANAR_U);
		src_pitch = src->GetPitch(PLANAR_U);
		dst_pitch = dst->GetPitch(PLANAR_U);
		rowsize = dst->GetRowSize(PLANAR_U);  
		height = dst->GetHeight(PLANAR_U);
//		ProcessFrame();
		Fieldcopy(dstp, srcp, rowsize,				// Just copy all chroma for now
				height, dst_pitch, src_pitch);

		// And the V plane, same sizes as U, different locations
		srcp = src->GetReadPtr(PLANAR_V);   
		dstp = dst->GetWritePtr(PLANAR_V);   
//		ProcessFrame();       
		Fieldcopy(dstp, srcp, rowsize,				// Just copy all chroma for now
				height, dst_pitch, src_pitch);

	}

	__asm emms;

    return dst;
}

int __stdcall UnFilter::ProcessFrame() 
{

	

    BOOL retc = 0;
	Fieldcopy(dstp, srcp, rowsize,				// copy first 2 rows
				2, dst_pitch, src_pitch);
	Fieldcopy(dstp+(height-2)*dst_pitch,
		srcp+(height-2)*src_pitch, rowsize, 
				2, dst_pitch, src_pitch);

	if (ItIsYV12)
	{
		if (SSEMMXenabled)
		{
			UnFilter_SSE_YV12( dstp + 2*dst_pitch, srcp + 2*src_pitch,
				rowsize,  height-4, 
				dst_pitch, src_pitch, 
				HSharp, VSharp);

		}
		else if (_3DNOWenabled)
		{
			UnFilter_3DNOW_YV12( dstp + 2*dst_pitch, srcp + 2*src_pitch,
				rowsize,  height-4, 
				dst_pitch, src_pitch, 
				HSharp, VSharp);
		}
		else
		{
			UnFilter_MMX_YV12( dstp + 2*dst_pitch, srcp + 2*src_pitch,
				rowsize,  height-4, 
				dst_pitch, src_pitch, 
				HSharp, VSharp);
		}
	}	
	else
	{
		if (SSEMMXenabled)
		{
			UnFilter_SSE( dstp + 2*dst_pitch, srcp + 2*src_pitch,
				rowsize,  height-4, 
				dst_pitch, src_pitch, 
				HSharp, VSharp);

		}
		else if (_3DNOWenabled)
		{
			UnFilter_3DNOW( dstp + 2*dst_pitch, srcp + 2*src_pitch,
				rowsize,  height-4, 
				dst_pitch, src_pitch, 
				HSharp, VSharp);
		}
		else
		{
			UnFilter_MMX( dstp + 2*dst_pitch, srcp + 2*src_pitch,
				rowsize,  height-4, 
				dst_pitch, src_pitch, 
				HSharp, VSharp);
		}
	}
	return 0;
}

int UnFilter::Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch)
{
BYTE* pDest = (BYTE*) dest;
BYTE* pSrc = (BYTE*) src;
	
	for (int i=0; i < rows; i++)
	{
		memcpy(pDest, pSrc, count);
		pSrc += src_pitch;
		pDest += dst_pitch;
	}
	return 0;
}

AVSValue __cdecl Create_UnFilter(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new UnFilter(args[0].AsClip(), args[1].AsInt(), args[2].AsInt(),
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("UnFilter", "cii", Create_UnFilter, 0);
    return "`UnFilter' DeInterlace";
}

