/////////////////////////////////////////////////////////////////////////////
// $Id: UnFilterAll.inc,v 1.0 2002/06/27 17:18:54 trbarry Exp $
/////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2002 Tom Barry  All rights reserved.
/////////////////////////////////////////////////////////////////////////////
//
//  This file is subject to the terms of the GNU General Public License as
//  published by the Free Software Foundation.  A copy of this license is
//  included with this software distribution in the file COPYING.  If you
//  do not have a copy, you may obtain a copy by writing to the Free
//  Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details
//
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//  See www.eff.org for details
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
// $Log: UnFilterAll.inc,v $
//
/////////////////////////////////////////////////////////////////////////////


// Add new Sharpen and Soften optimized to reverse previous  filtering
// Some pseudo-rationale:
/*
Assume that someone vertically filtered the video for interlace using a simple
center weighted moving average, creating new pixels Z[k] from old pixels X[k]
using a weighting factor w (0 < w < 1, w close to 1). Then we might be seeing:

1)	Z[k] = w X[k] + .5 (1-w) (X[k-1] + X[k+1])

useful abbrevs to avoid typing: 

Z[k-2] == Zi, X[k-2] == Xi
Z[k-1] == Zj, etc
Z[k] == Zk, X[k] == Xk
Z[k+1] == Zl, etc
Z[k+2] == Zm, etc

Q == .5 (1-w) / w, will need later         # Q = .5 w=.5


so:

2) Zk = w Xk + .5 (1-w) (Xj + Xl)           # Zk = .5 3 + .5 .5 (2 +2) = 2.5

But, if we wanted to solve for Xk

3) Xk =  [ Zk - .5 (1-w) (Xj + Xl) ] / w     # (2.5 - .25 (2+2))) / .5 = 3
	  
4) Xk =   Zk / w - Q (Xj + Xl)               # 2.5 / .5 - .5 (2+2) = 3

5) Xj =   Zj / w - Q (Xi + Xk)               # 2 / .5 - .5 (1+3) = 2 

6) Xl =   Zl / w - Q (Xk + Xm)               # 2 / .5 - .5 (3+1) = 2

and substituting for Xj and Xl in 4) from 5) and 6)

7) Xk =   Zk / w - Q [ Zj / w - Q (Xi + Xk) + Zl / w - Q (Xk + Xm) ]

								# 2.5/.5 - .5 [2/.5 - .5(1+3) + 2/.5 - .5(3+1)]
								# 5 - .5[4 - 2 + 4 - 2] = 5 - .5[4] = 3

rearranging a bit

8) Xk =  Zk / w - Q [ (Zj + Zl) / w  - Q (Xi + Xm + 2 Xk) ]
								
								# 2.5/.5 - .5 [(2+2)/.5 - .5 (1 + 1 + 2 3)]
								# 5 - .5 [8 - .5 (8)] = 5 - .5 [4] = 3
								
9) Xk =  Zk / w - Q (Zj + Zl) / w + Q^2 (Xi + Xm) + 2 Q^2 Xk

								# 2.5/.5 - .5 (2+2)/.5 + .25 (1+1) + 2 .25 3
								# 5 - 4 + .5 + 1.5 = 3

moving all Xk terms to the left

10) (1 - 2 Q^2) Xk = Zk / w - Q (Zj + Zl) / w + Q^2 (Xi + Xm)

								# (1 - 2 .25)3 = 2.5/.5 - .5(2+2)/.5 + .25(1+1)
								# (.5)3 = 5 - 4 + .5

11) Xk = [ Zk / w - Q (Zj + Zl) / w + Q^2 (Xi + Xm) ] / (1 - 2 Q^2)

								# [2.5/.5 - .5(2+2)/.5 + .25(1+1)] / (1 - 2 .25)
								# [5 - 4 + .5] / .5 = 1.5/.5 = 3

12) Xk = Zk - Q (Zj + Zl) + w Q^2 (Xi + Xm)    
		 ---------------------------------
		    	w (1 - 2 Q^2)  
				
								# [2.5 - .5(2+2) + .5 .25 (1+1)] / [.5 (1 - 2 .25)}
								# [2.5 - 2 + .25] / .5 (.5)
								# .75 / .25 = 3

but we'd like it in this form

13) Xk = A Zk - B avg(Zj,Zl) + C avg(Xi,Xm)

where:

14) A = 1 / [w (1 - 2 Q^2)]		# 1 / [.5 (1 - 2 .25)] = 1 / (.5 .5) = 4

15) B = 2 Q / [w (1 - 2 Q^2)]   # 2 .5 / [.5 (1 - 2 .25)] = 1 / .25 = 4

16( C = 2 Q^2 / [1 - 2 Q^2]		# 2 .25 / [1 - 2 .25] = .5 / .5 = 1

		# from 13)  3 =  4 2.5 - 4 2 + 1 1 = 10 - 8 + 1 = 3

Note we still don't know the orig values of Xi but we are out of patience, CPU,
and math ability so we just assume Xi=Zi for the remaining terms. Hopefully 
it will be close. Note the powers are increasing and the values are < 1. Also
as set up even the third term goes to 0 somewhere with a parm value in the 20's.

*/


#include "UnFilter.h"

static BOOL  HSharpVSharp( BYTE* Dest, const BYTE* Source,
				 int rowsize, int src_pitch,
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC);

static BOOL  HSharpVSoft( BYTE* Dest, const BYTE* Source,
				 int rowsize, int src_pitch,
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC);

static BOOL  HSoftVSharp( BYTE* Dest, const BYTE* Source,
				 int rowsize, int src_pitch,
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC);

static BOOL  HSoftVSoft( BYTE* Dest, const BYTE* Source,
				 int rowsize, int src_pitch,
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC);


BOOL FUNCT_NAME(BYTE* pDest, const BYTE* pSrc,  
			int rowsize,  int height,
			int dst_pitch, int src_pitch,
			int HSharp, int VSharp)
{
    __int64 QHA;
    __int64 QHB;
    __int64 QHC; 

    __int64 QVA;
    __int64 QVB;
    __int64 QVC; 

	// fill in a few parm values where other rtns and asm can access them


	// Calc values for Horizontal
    __int64 i;
    int w = (HSharp > 0)                 // note-adj down for overflow
            ? 1000 - (HSharp * 38 / 10)  // overflow, use 38%, 0<w<1000                                             
            : 1000 - (HSharp * 150 / 10); // bias towards workable range 
    int Q = 500 * (1000 - w) / w;                      // Q as 0 - 1K, max 16k        
    int Q2 = (Q*Q) / 1000;                             // Q^2 as 0 - 1k
    int denom = (w * (1000 - 2 * Q2)) / 1000;          // [w (1-2q^2)] as 0 - 1k    
    int HA = 64000 / denom;                             // HA as 0 - 64
    int HB = 128 * Q / denom;                           // HB as 0 - 64
    int HC = 64 - HA + HB;                                // so HA-HB+HC=64, unbiased weight
	int VA;
	int VB;
	int VC;

	int line;
    
	i = HA;                          
    QHA = i << 48 | i << 32 | i << 16 | i;

    i = HC;
    QHC = i << 48 | i << 32 | i << 16 | i;

	// Calc values for Vertical
    w = (VSharp > 0)                 // note-adj down for overflow
            ? 1000 - (VSharp * 38 / 10)  // overflow, use 38%, 0<w<1000 
                                                      
            : 1000 - (VSharp * 150 / 10); // bias towards workable range 
    Q = 500 * (1000 - w) / w;                      // Q as 0 - 1K, max 16k        
    Q2 = (Q*Q) / 1000;                             // Q^2 as 0 - 1k
    denom = (w * (1000 - 2 * Q2)) / 1000;          // [w (1-2q^2)] as 0 - 1k    
    VA = 64000 / denom;                             // VA as 0 - 64
    VB = 128 * Q / denom;                           // VB as 0 - 64
//    VC =  ((w * Q2) / (denom*500)) ;                // VC as 0 - 64

    VC = 64 - VA + VB;

/* for debug
    if ((VA-VB+VC-64 > 1) || (VA-VB+VC-64 < -1))
    {
        VC = 64 - VA + VB;
    }

    // for too large plus minus Sharpness
    if (VA < 32 || VA > 64 || VB > 0 || VB < -64 || VC < 0 || VC > 64 || w < 0 )
    {
        VA = VA;
    }
    // for too large plus Sharpness
    if (VA > 127 || VB > 127 || VC > 127 || VA < 64 || VB < 0 || VC < 0)
    {
        VC = VC;  // for setting stops
    }  
*/


   i = VA;                          
    QVA = i << 48 | i << 32 | i << 16 | i;
    i = VC;
    QVC = i << 48 | i << 32 | i << 16 | i;

    if (VB < 0)
    {					// Vertical is Soft
        i = -VB;
        QVB = i << 48 | i << 32 | i << 16 | i;

		if (HB < 0)		// Horiz is Soft
		{
			i = -HB;
			QHB = i << 48 | i << 32 | i << 16 | i;

			for (line = 0; line < height; ++line)
			{
				HSoftVSoft(pDest + line*dst_pitch, pSrc + line*src_pitch,
					src_pitch, rowsize, 
					QHA, QHB, QHC, QVA, QVB, QVC);
			}

		}
		else
		{
			i = HB;		// Horiz is Sharp
			QHB = i << 48 | i << 32 | i << 16 | i;

			for (line = 0; line < height; ++line)
			{
				HSharpVSoft(pDest + line*dst_pitch, pSrc + line*src_pitch,
					src_pitch, rowsize, 
					QHA, QHB, QHC, QVA, QVB, QVC);
			}
		}
    }

    else
  
	 {					// Vertical is Sharp
        i = VB;
        QVB = i << 48 | i << 32 | i << 16 | i;

		if (HB < 0)		// Horiz is Soft
		{
			i = -HB;
			QHB = i << 48 | i << 32 | i << 16 | i;

			for (line = 0; line < height; ++line)
			{
				HSoftVSharp(pDest + line*dst_pitch, pSrc + line*src_pitch,
					src_pitch, rowsize, 
					QHA, QHB, QHC, QVA, QVB, QVC);
			}

		}
		else
		{
			i = HB;		// Horiz is Sharp
			QHB = i << 48 | i << 32 | i << 16 | i;

			for (line = 0; line < height; ++line)
			{
				HSharpVSharp(pDest + line*dst_pitch, pSrc + line*src_pitch,
					src_pitch, rowsize, 
					QHA, QHB, QHC, QVA, QVB, QVC);
			}
		}
	}

 	__asm emms;

    return 0;
}

    

BOOL  HSharpVSharp( BYTE* Dest, const BYTE* Source,			// Sharpend H & V
				int src_pitch, int rowsize, 
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC)
{
#include "DoSoftSharp.inc"
}


#define REALLY_HORIZ_SOFTNESS    
BOOL  HSoftVSharp( BYTE* Dest, const BYTE* Source,			// Soften H & Sharpen V
				int src_pitch, int rowsize, 
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC)
{
#include "DoSoftSharp.inc"
}
#undef REALLY_HORIZ_SOFTNESS    


#define REALLY_VERT_SOFTNESS    
BOOL  HSharpVSoft( BYTE* Dest, const BYTE* Source,			// Sharpen H & Soften V
				int src_pitch, int rowsize, 
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC)
{
#include "DoSoftSharp.inc"
}


#define REALLY_HORIZ_SOFTNESS    
BOOL  HSoftVSoft( BYTE* Dest, const BYTE* Source,			// Soften H & V
				int src_pitch, int rowsize, 
				__int64 QHA, __int64 QHB, __int64 QHC, 
				__int64 QVA, __int64 QVB, __int64 QVC)

{
#include "DoSoftSharp.inc"
}
